@@ -1,6 +1,9 @@
 library(testthat)
 library(rmumps)
 library(Matrix)
-library(slam)
+library(methods)
+if (getRversion() >= "3.4.0") {
+  library(slam)
+}
 
 test_check("rmumps")
